@@ -1,5 +1,4 @@
 use std::any::TypeId;
-use std::cell::{Ref, RefMut};
 use std::iter::FusedIterator;
 use std::marker::PhantomData;
 use std::mem::{transmute, ManuallyDrop};
@@ -7,7 +6,7 @@ use std::ptr::NonNull;
 use std::slice::Iter;
 
 use crate::{
-    archetype::Archetype,
+    archetype::{Archetype, Ref, RefMut},
     world::{Entity, EntityMetadata, World},
 };
 
@@ -25,6 +24,7 @@ use crate::rayon::QueryParIter;
 /// * `Option<Q>` - optional component(s)
 /// * `With<Q, C>` to filter `Q` for presence of `C`
 /// * `Without<Q, C>` to filter `Q` for absence of `C`
+/// * `Matches<Q> to indicate which entities match `Q`
 ///
 /// Note that [Entities](Entity) are components themselves, so they can be optionally obtained in a query,
 /// like `Query<Entity, &C, &mut D>`.
@@ -96,9 +96,9 @@ where
     S: QuerySpec,
 {
     tag_gen: (u32, u16),
-    types: Vec<(u16, <S::Fetch as Fetch<'static>>::Ty)>,
+    types: Box<[(u16, <S::Fetch as Fetch<'static>>::Ty)]>,
     refs: Vec<ManuallyDrop<<S::Fetch as Fetch<'static>>::Ref>>,
-    ptrs: Vec<Option<<S::Fetch as Fetch<'static>>::Ptr>>,
+    ptrs: Box<[Option<<S::Fetch as Fetch<'static>>::Ptr>]>,
 }
 
 #[allow(clippy::non_send_fields_in_send_ty)]
@@ -158,75 +158,38 @@ where
 
         self.refs.clear();
 
-        let types: &'w [(u16, <S::Fetch as Fetch<'w>>::Ty)] = unsafe { transmute(&*self.types) };
-        let refs: &'w mut Vec<<S::Fetch as Fetch<'w>>::Ref> = unsafe { transmute(&mut self.refs) };
-        let ptrs: &'w mut Vec<Option<<S::Fetch as Fetch<'w>>::Ptr>> =
-            unsafe { transmute(&mut self.ptrs) };
+        let ref_ = QueryRef {
+            world,
+            types: unsafe { transmute(&*self.types) },
+            refs: unsafe { transmute(&mut self.refs) },
+            ptrs: unsafe { transmute(&mut *self.ptrs) },
+        };
 
-        for (idx, ty) in types {
+        for (idx, ty) in ref_.types {
             let archetype = &world.archetypes[*idx as usize];
 
             if archetype.len() != 0 {
-                refs.push(unsafe { S::Fetch::borrow(archetype, *ty) });
+                ref_.refs.push(unsafe { S::Fetch::borrow(archetype, *ty) });
             }
         }
 
-        QueryRef {
-            world,
-            types,
-            refs,
-            ptrs,
-        }
+        ref_
     }
 
     #[cold]
     #[inline(never)]
     fn find(&mut self, world: &World) {
-        self.types.clear();
+        self.types = world
+            .archetypes
+            .iter()
+            .enumerate()
+            .filter_map(|(idx, archetype)| S::Fetch::find(archetype).map(|ty| (idx as u16, ty)))
+            .collect();
 
-        for (idx, archetype) in world.archetypes.iter().enumerate() {
-            if let Some(ty) = S::Fetch::find(archetype) {
-                self.types.push((idx as u16, ty));
-            }
-        }
+        self.ptrs = world.archetypes.iter().map(|_| None).collect();
 
         self.tag_gen = world.tag_gen();
     }
-
-    /// Narrow down a query to entities that have a certain component,
-    /// without borrowing that component.
-    ///
-    /// For use with prepared queries, see [With].
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// # use rs_ecs::*;
-    /// let query = Query::<(&i32,)>::new().with::<bool>();
-    /// ```
-    pub fn with<C>(self) -> Query<With<S, C>>
-    where
-        C: 'static,
-    {
-        Query::new()
-    }
-
-    /// Narrow down a query to entities that do not have a certain component.
-    ///
-    /// For use with prepared queries, see [Without].
-    ///
-    /// # Examples
-    ///
-    /// ```
-    /// # use rs_ecs::*;
-    /// let query = Query::<(&i32,)>::new().without::<bool>();
-    /// ```
-    pub fn without<C>(self) -> Query<Without<S, C>>
-    where
-        C: 'static,
-    {
-        Query::new()
-    }
 }
 
 /// Borrow of the [World] for a [Query]. Required to obtain an iterator.
@@ -237,7 +200,7 @@ where
     world: &'w World,
     types: &'w [(u16, <S::Fetch as Fetch<'w>>::Ty)],
     refs: &'w mut Vec<<S::Fetch as Fetch<'w>>::Ref>,
-    ptrs: &'w mut Vec<Option<<S::Fetch as Fetch<'w>>::Ptr>>,
+    ptrs: &'w mut [Option<<S::Fetch as Fetch<'w>>::Ptr>],
 }
 
 impl<S> QueryRef<'_, S>
@@ -314,11 +277,10 @@ where
     pub fn map<'q>(&'q mut self) -> QueryMap<'q, S> {
         let types: &'q [(u16, <S::Fetch as Fetch<'q>>::Ty)] = unsafe { transmute(self.types) };
 
-        let ptrs: &'q mut Vec<Option<<S::Fetch as Fetch<'q>>::Ptr>> =
+        let ptrs: &'q mut [Option<<S::Fetch as Fetch<'q>>::Ptr>] =
             unsafe { transmute(&mut *self.ptrs) };
 
-        ptrs.clear();
-        ptrs.resize(self.world.archetypes.len(), None);
+        ptrs.fill(None);
 
         for (idx, ty) in types {
             let archetype = &self.world.archetypes[*idx as usize];
@@ -505,7 +467,7 @@ where
     C: 'static,
 {
     type Ty = u16;
-    type Ref = Ref<'q, ()>;
+    type Ref = Ref<'q>;
     type Ptr = NonNull<C>;
 
     type Item = &'q C;
@@ -547,7 +509,7 @@ where
     C: 'static,
 {
     type Ty = u16;
-    type Ref = RefMut<'q, ()>;
+    type Ref = RefMut<'q>;
     type Ptr = NonNull<C>;
 
     type Item = &'q mut C;
@@ -624,8 +586,6 @@ unsafe impl<F> FetchShared for TryFetch<F> where F: FetchShared {}
 /// A query specification to iterate over entities with a certain component,
 /// but without borrowing that component.
 ///
-/// See also [Query::with()]
-///
 /// # Examples
 ///
 /// A query for components of type `u32` and `bool`,
@@ -686,8 +646,6 @@ unsafe impl<F, C> FetchShared for FetchWith<F, C> where F: FetchShared {}
 
 /// A query specification to iterate over entities without a certain component.
 ///
-/// See also [Query::without()]
-///
 /// # Examples
 ///
 /// A query for components of type `u32` and `bool`,
@@ -746,6 +704,75 @@ where
 
 unsafe impl<F, C> FetchShared for FetchWithout<F, C> where F: FetchShared {}
 
+/// A query specification to indicate which entities match the inner query,
+/// but without borrowing any components.
+///
+/// # Examples
+///
+/// ```
+/// # use rs_ecs::*;
+/// let mut world = World::new();
+///
+/// let entity1 = world.alloc();
+/// world.insert(entity1, (42_i32, 1.0_f32));
+///
+/// let entity2 = world.alloc();
+/// world.insert(entity2, (23_i32,));
+///
+/// let mut query = Query::<(&i32, Matches<&f32>)>::new();
+/// let mut query = query.borrow(&world);
+/// let mut query = query.map();
+///
+/// let (i1, f1) = query.get(entity1).unwrap();
+/// assert_eq!(*i1, 42);
+/// assert!(f1);
+///
+/// let (i2, f2) = query.get(entity2).unwrap();
+/// assert_eq!(*i2, 23);
+/// assert!(!f2);
+/// ```
+pub struct Matches<S>(PhantomData<S>);
+
+impl<S> QuerySpec for Matches<S>
+where
+    S: QuerySpec,
+{
+    type Fetch = FetchMatches<S::Fetch>;
+}
+
+pub struct FetchMatches<F>(PhantomData<F>);
+
+unsafe impl<'q, F> Fetch<'q> for FetchMatches<F>
+where
+    F: Fetch<'q>,
+{
+    type Ty = bool;
+    type Ref = ();
+    type Ptr = bool;
+
+    type Item = bool;
+
+    fn find(archetype: &Archetype) -> Option<Self::Ty> {
+        Some(F::find(archetype).is_some())
+    }
+
+    unsafe fn borrow(_archetype: &'q Archetype, _ty: Self::Ty) -> Self::Ref {}
+
+    unsafe fn base_pointer(_archetype: &'q Archetype, ty: Self::Ty) -> Self::Ptr {
+        ty
+    }
+
+    fn dangling() -> Self::Ptr {
+        false
+    }
+
+    unsafe fn deref(ptr: Self::Ptr, _idx: u32) -> Self::Item {
+        ptr
+    }
+}
+
+unsafe impl<F> FetchShared for FetchMatches<F> {}
+
 macro_rules! impl_fetch_for_tuples {
     () => {};
 
@@ -820,6 +847,7 @@ mod tests {
     use super::*;
 
     use std::mem::{forget, size_of};
+    use std::panic::{catch_unwind, AssertUnwindSafe};
 
     fn spawn_three(world: &mut World) {
         let ent = world.alloc();
@@ -923,6 +951,23 @@ mod tests {
         let _ = query.borrow(&world);
     }
 
+    #[test]
+    fn conflicting_borrow_leaves_world_in_consistent_state() {
+        let mut world = World::new();
+
+        spawn_three(&mut world);
+
+        let res = catch_unwind(AssertUnwindSafe(|| {
+            let mut query = Query::<(&i32, Option<(&mut i32, &bool)>)>::new();
+            let _ = query.borrow(&world);
+        }));
+
+        assert!(res.is_err());
+
+        let mut query = Query::<&mut i32>::new();
+        let _ = query.borrow(&world);
+    }
+
     #[test]
     fn borrows_can_be_reused() {
         let mut world = World::new();
@@ -982,7 +1027,7 @@ mod tests {
 
         spawn_three(&mut world);
 
-        let mut query = Query::<&i32>::new().with::<bool>();
+        let mut query = Query::<With<&i32, bool>>::new();
         let sum = query.borrow(&world).iter().sum::<i32>();
 
         assert_eq!(sum, 42);
@@ -994,7 +1039,7 @@ mod tests {
 
         spawn_three(&mut world);
 
-        let mut query = Query::<&i32>::new().without::<bool>();
+        let mut query = Query::<Without<&i32, bool>>::new();
         let sum = query.borrow(&world).iter().sum::<i32>();
 
         assert_eq!(sum, 23 + 1);
